@@ -20,9 +20,10 @@
 
 #include <algorithm>
 
-#include <bctoolbox/defs.h>
+#include "bctoolbox/defs.h"
 
 #include "ortp/logging.h"
+#include "ortp/ortp.h"
 #include "ortp/rtpsession.h"
 #include "rtpbundle.h"
 #include "rtpsession_priv.h"
@@ -73,14 +74,12 @@ extern "C" void rtp_bundle_set_primary_session(RtpBundle *bundle, RtpSession *se
 	reinterpret_cast<RtpBundleCxx *>(bundle)->setPrimarySession(session);
 }
 
-extern "C" char *rtp_bundle_get_session_mid(RtpBundle *bundle, RtpSession *session) {
-	try {
-		auto &mid = reinterpret_cast<RtpBundleCxx *>(bundle)->getSessionMid(session);
-		return bctbx_strdup(mid.c_str());
-	} catch (std::string const &e) {
-		ortp_warning("RtpBundle[%p]: cannot get mid for session (%p): %s", bundle, session, e.c_str());
-		return nullptr;
-	}
+extern "C" char *rtp_session_get_mid(RtpSession *session) {
+	char *ret_value = NULL;
+	bctbx_mutex_lock(&session->main_mutex);
+	if (session->mid) ret_value = bctbx_strdup(session->mid);
+	bctbx_mutex_unlock(&session->main_mutex);
+	return ret_value;
 }
 
 extern "C" mblk_t *rtp_bundle_dispatch(RtpBundle *bundle, bool_t is_rtp, mblk_t *m) {
@@ -104,13 +103,7 @@ void RtpBundleCxx::setMidId(int id) {
 	mMidId = id;
 }
 
-void RtpBundleCxx::addSession(const std::string &mid, RtpSession *session) {
-	const std::lock_guard guard(mAssignmentMutex);
-
-	// Search for the session in both maps to check if it hasn't already been inserted.
-	if (findSession(session))
-		ortp_error("RtpBundle[%p]: Cannot add session (%p) as it is already in the bundle", this, session);
-
+void RtpBundleCxx::enterSession(RtpSession *session, const std::string &mid) {
 	// Check for the mode. If SENDONLY, we already know it's SSRC, and we can assign it now.
 	// Otherwise, add it in the waiting for assignment map.
 	switch (session->mode) {
@@ -125,13 +118,23 @@ void RtpBundleCxx::addSession(const std::string &mid, RtpSession *session) {
 			mWaitingForAssignment.emplace(mid, session);
 			break;
 	}
+}
+
+void RtpBundleCxx::addSession(const std::string &mid, RtpSession *session) {
+	const std::lock_guard guard(mAssignmentMutex);
+
+	// Search for the session in both maps to check if it hasn't already been inserted.
+	if (findSession(session))
+		ortp_error("RtpBundle[%p]: Cannot add session (%p) as it is already in the bundle", this, session);
+
+	enterSession(session, mid);
 
 	if (!mPrimary) {
 		mPrimary = session;
 		session->is_primary = TRUE;
 	}
 
-	rtp_session_set_bundle(session, reinterpret_cast<RtpBundle *>(this));
+	rtp_session_set_bundle(session, reinterpret_cast<RtpBundle *>(this), mid.c_str());
 }
 
 bool RtpBundleCxx::findSession(RtpSession *session) const {
@@ -234,7 +237,7 @@ void RtpBundleCxx::removeSession(RtpSession *session) {
 }
 
 void RtpBundleCxx::clearSession(RtpSession *session) {
-	rtp_session_set_bundle(session, nullptr);
+	rtp_session_set_bundle(session, nullptr, nullptr);
 
 	if (mPrimary != nullptr) {
 		if (session == mPrimary) {
@@ -250,15 +253,17 @@ void RtpBundleCxx::clearSession(RtpSession *session) {
 	}
 }
 
-void RtpBundleCxx::clear() {
-	for (const auto &entry : mSsrcToSession) {
-		entry.second.rtpSession->bundle = nullptr;
+void RtpBundleCxx::clearFromWaitingMap(RtpSession *session, const std::string &mid) {
+	const auto [first, last] = mWaitingForAssignment.equal_range(mid);
+	for (auto s = first; s != last;) {
+		if (s->second == session) {
+			s = mWaitingForAssignment.erase(s);
+		} else ++s;
 	}
-	mSsrcToSession.clear();
+}
 
-	for (const auto &entry : mWaitingForAssignment) {
-		entry.second->bundle = nullptr;
-	}
+void RtpBundleCxx::clear() {
+	mSsrcToSession.clear();
 	mWaitingForAssignment.clear();
 
 	mSsrcToMid.clear();
@@ -332,26 +337,6 @@ void RtpBundleCxx::setPrimarySession(RtpSession *session) {
 	}
 }
 
-const std::string &RtpBundleCxx::getSessionMid(RtpSession *session) const {
-	const auto inMainMap = std::find_if(
-	    mSsrcToSession.begin(), mSsrcToSession.end(),
-	    [session](const std::pair<uint32_t, BundleSession> &t) -> bool { return t.second.rtpSession == session; });
-
-	if (inMainMap != mSsrcToSession.end()) {
-		return inMainMap->second.mid.mid;
-	}
-
-	const auto inWaitingMap =
-	    std::find_if(mWaitingForAssignment.begin(), mWaitingForAssignment.end(),
-	                 [session](const std::pair<std::string, RtpSession *> &t) -> bool { return t.second == session; });
-
-	if (inWaitingMap != mWaitingForAssignment.end()) {
-		return inWaitingMap->first;
-	}
-
-	throw std::string("the session must be in the bundle!");
-}
-
 static void getSsrcFromSdes(void *userData,
                             uint32_t ssrc,
                             rtcp_sdes_type_t t,
@@ -504,6 +489,8 @@ void RtpBundleCxx::updateBundleSession(BundleSession &session, const std::string
 	if (!mid.empty() && mid != session.mid.mid && sequenceNumber > session.mid.sequenceNumber) {
 		session.mid.mid = mid;
 		session.mid.sequenceNumber = sequenceNumber;
+		/* reflect the change into the RtpSession's own mid attribute. */
+		rtp_session_set_bundle(session.rtpSession, reinterpret_cast<RtpBundle *>(this), mid.c_str());
 	}
 }
 
@@ -531,23 +518,50 @@ RtpSession *RtpBundleCxx::checkForSession(const mblk_t *m, bool isRtp, bool isOu
 		}
 	}
 
-	// Try to route the packet to the correct session.
-	if (const auto it = mSsrcToSession.find(ssrc); it != mSsrcToSession.end()) {
-		updateBundleSession(it->second, mid, isRtp ? rtp_get_seqnumber(m) : 0);
-		return it->second.rtpSession;
+	// Try to route the packet to the correct session by SSRC
+	auto bundleSessionIt = mSsrcToSession.find(ssrc);
+
+	if (bundleSessionIt != mSsrcToSession.end() && isRtp) {
+		/* The SSRC is known.
+		 * Detect when an incoming SSRC gets its mid changed.
+		 * For outgoing sessions, we don't care.
+		 */
+		const std::string &previouslyKnownMid = bundleSessionIt->second.mid.mid;
+		if (!mid.empty() && !previouslyKnownMid.empty() && previouslyKnownMid != mid) {
+			ortp_message("RtpBundle[%p]: SSRC %u moved from mid %s to %s", this, ssrc, previouslyKnownMid.c_str(),
+			             mid.c_str());
+			if (!isOutgoing) {
+				/*
+				 * We need to clear everything for this SSRC, so that everything restarts as it was new, and the packets
+				 * be routed to the correct media stream.
+				 */
+				mSsrcToSession.erase(ssrc);
+				mSsrcToMid.erase(ssrc);
+				bundleSessionIt = mSsrcToSession.end();
+			}
+			/* For outgoing stream, we don't care: let's reuse the former RtpSession
+			 * to ensure continuity of sequence numbers.
+			 */
+		}
 	}
 
+	if (bundleSessionIt != mSsrcToSession.end()) {
+		updateBundleSession(bundleSessionIt->second, mid, isRtp ? rtp_get_seqnumber(m) : 0);
+		return bundleSessionIt->second.rtpSession;
+	}
+
+	auto ssrcToMidIt = mSsrcToMid.find(ssrc);
 	// Retrieve or update the MID from the association map.
-	if (const auto it = mSsrcToMid.find(ssrc); mid.empty()) {
+	if (mid.empty()) {
 		// If there is no mid in the packet, check if we have it stored for this ssrc.
-		if (it == mSsrcToMid.end()) {
+		if (ssrcToMidIt == mSsrcToMid.end()) {
 			ortp_warning("RtpBundle[%p]: Packet with SSRC %u doesn't have any mid and no corresponding mid in bundle",
 			             this, ssrc);
 			return nullptr;
 		}
 
-		mid = it->second;
-	} else if (it == mSsrcToMid.end()) {
+		mid = ssrcToMidIt->second;
+	} else if (ssrcToMidIt == mSsrcToMid.end()) {
 		// We have a mid in the packet, but not in the map. Insert it.
 		mSsrcToMid[ssrc] = mid;
 	}
@@ -562,8 +576,9 @@ RtpSession *RtpBundleCxx::checkForSession(const mblk_t *m, bool isRtp, bool isOu
 
 				// Check if this blank session knows the payload type of the incoming packet.
 				if (session->rcv.pt == rtp_get_payload_type(m)) {
-					ortp_message("RtpBundle[%p]: Assigning incoming SSRC %u to session %p using RTP with pt %d", this,
-					             ssrc, session, rtp_get_payload_type(m));
+					ortp_message(
+					    "RtpBundle[%p]: Assigning incoming SSRC %u with mid %s to session %p using RTP with pt %d",
+					    this, ssrc, mid.c_str(), session, rtp_get_payload_type(m));
 
 					session->ssrc_set = TRUE;
 					session->rcv.ssrc = ssrc;
@@ -571,7 +586,6 @@ RtpSession *RtpBundleCxx::checkForSession(const mblk_t *m, bool isRtp, bool isOu
 					// Assign the session to the incoming ssrc and remove this session from the assignment map.
 					mSsrcToSession.emplace(ssrc, BundleSession{{mid, 0}, session});
 					mWaitingForAssignment.erase(s);
-
 					return session;
 				}
 			}
@@ -627,7 +641,17 @@ RtpSession *RtpBundleCxx::checkForSession(const mblk_t *m, bool isRtp, bool isOu
 			mSsrcToSession.emplace(isOutgoing ? newRtpSession->snd.ssrc : newRtpSession->rcv.ssrc,
 			                       BundleSession{{mid, 0}, newRtpSession});
 			if (newRtpSession->bundle == nullptr)
-				rtp_session_set_bundle(newRtpSession, reinterpret_cast<RtpBundle *>(this));
+				rtp_session_set_bundle(newRtpSession, reinterpret_cast<RtpBundle *>(this), mid.c_str());
+		}
+	}
+	if (newRtpSession == nullptr) {
+		if (isRtp) {
+			ortp_warning(
+			    "RtpBundle[%p]: don't know what to do with received RTP packet with mid: %s SSRC: %u payload-type: %i)",
+			    this, mid.c_str(), ssrc, rtp_get_payload_type(m));
+		} else {
+			ortp_warning("RtpBundle[%p]: don't know what to do with received RTCP packet with mid: %s.", this,
+			             mid.c_str());
 		}
 	}
 
@@ -644,6 +668,8 @@ std::optional<mblk_t *> RtpBundleCxx::dispatchRtpMessage(mblk_t *m) {
 	RtpSession *session = checkForSession(m, true);
 	if (session == nullptr) {
 		freemsg(m);
+		mPrimary->stats.discarded++;
+		ortp_global_stats.discarded++;
 		return {};
 	}
 
